/*
 * Copyright (c) 2008-2010 by Gerasimenko Roman.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistribution of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistribution in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *
 * 3. The name "JFX" must not be used to endorse or promote
 *     products derived from this software without prior written
 *     permission.
 *     For written permission, please contact roman.gerasimenko@gmail.com
 *
 * 4. Products derived from this software may not be called "JFX",
 *     nor may "JFX" appear in their name, without prior written
 *     permission of Gerasimenko Roman.
 *
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE JFX CONTRIBUTORS
 *  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

import com.jfx.*;
import com.jfx.net.JFXServer;
import com.jfx.strategy.Strategy;
import com.jfx.strategy.StrategyRunner;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Enumeration;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.InetAddress;

public class JFXExample extends Strategy {
    static {
        String publicIP = getPublicIP();
        System.out.println("Using public IP: " + publicIP);
        System.setProperty("jfx_server_host", publicIP);
//        System.setProperty("jfx_activation_key", "0");
//        System.setProperty("jfx_server_port", "7799");
    }

    private static String getPublicIP() {
        try {
            String bindAddress = null;
            for (Enumeration ifaces = NetworkInterface.getNetworkInterfaces(); ifaces.hasMoreElements();) {
                NetworkInterface iface = (NetworkInterface) ifaces.nextElement();
                boolean ifPrinted = false;
                for (Enumeration ips = iface.getInetAddresses(); ips.hasMoreElements();) {
                    InetAddress ia = (InetAddress) ips.nextElement();
                    String address = ia.getHostAddress();
                    boolean skip = address.startsWith("192.168.") || address.startsWith("127.0.0.1") || address.indexOf(':') > 0;
                    if (!skip) {
                        if (!ifPrinted) {
                            System.out.println("Interface: " + iface.getDisplayName() + '/' + iface.getName());
                            ifPrinted = true;
                        }
                        System.out.println("    " + ia.getCanonicalHostName() + ": " + address);
//                        System.out.println("    SiteLocal=" + ia.isSiteLocalAddress());
                        if (!ia.isSiteLocalAddress()) {
                            return address;
                        } else {
                            bindAddress = address;
                        }
                    }
                }
            }
            return bindAddress;
        } catch (SocketException e) {
            throw new RuntimeException(e);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }
    //
    private int shift;
    public double sellPrice;
    private boolean laTestDone = false;
//    private static final String EURUSD = "EURUSD";

    public JFXExample() {
    }

    public synchronized void init(String symbol, int period, StrategyRunner strategyRunner) throws ErrUnknownSymbol, IOException {
        super.init(symbol, period, strategyRunner);
        //
        ArrayList<Instrument> instruments = getInstruments();
        if (instruments != null) {
            System.out.println("Available instruments:");
            for (int i = 0; i < instruments.size(); i++) {
                Instrument s = instruments.get(i);
                System.out.println("" + s);
            }
            System.out.println("");
        }
    }

    boolean infoPrinted;

    //final static long[] CNT_THRESHOLDS = new long[]{50, 100, 250, 500, 10000};
    final static long[] CNT_THRESHOLDS = new long[]{5, 10, 25, 50, 100};

    private void getCounted(long createdRespDuration, long[] createCnt) {
        for (int i = 0; i < CNT_THRESHOLDS.length; i++) {
            long cntThreshold = CNT_THRESHOLDS[i];
            if (cntThreshold > createdRespDuration) {
                createCnt[i]++;
                break;
            }
        }
    }

    public static String lpad(String s, int len, char pad) {
        int padLen = len - s.length();
        if (padLen > 0) {
            char[] p = new char[padLen];
            for (int i = 0; i < padLen; ++i) {
                p[i] = pad;
            }
            return (new String(p)).concat(s);
        } else {
            return s;
        }
    }

    private String format(String ts, String s, String s1, String s2, String s3, String s4, long[] timeCounts) {
        int width = 20;
        StringBuffer msg = new StringBuffer(ts + s + lpad(" ", width - s.length(), ' ') +
                s1 + lpad(" ", width - s1.length(), ' ') +
                s2 + lpad(" ", width - s2.length(), ' ') +
                s3 + lpad(" ", width - s3.length(), ' ') +
                s4 + lpad(" ", width - s4.length(), ' ') + "cnt: ");
        for (int i = 0; i < timeCounts.length; i++) {
            long timeCount = timeCounts[i];
            String str = String.valueOf(timeCount);
            msg.append(str).append(lpad(" ", 5 - str.length(), ' '));
        }
        return msg.toString();
    }

    public void coordinate() {
        try {
            //refreshRates();
            //double _point = marketInfo(getSymbol(), MarketInfo.MODE_POINT);
            //double _bid = marketInfo(getSymbol(), MarketInfo.MODE_BID);
            //double _ask = marketInfo(getSymbol(), MarketInfo.MODE_ASK);
            //System.out.println("point=[" + _point + "] sym=" + getSymbol());
            //System.out.println("bid=[" + _bid + "] [" + (Math.round(_bid/_point) * _point) + ']');
            //System.out.println("bid=[" + _bid + "] [" + (Math.round(_bid * 1000000) / 1000000.0) + ']');
            //System.out.println("ask=[" + _ask + "] [" + (Math.round(_ask/_point) * _point) + ']');
            //
            if (!laTestDone) {
                laTestDone = true;
                long[] createCnt = new long[CNT_THRESHOLDS.length];
                long createMax = Long.MIN_VALUE,
                        createMin = Long.MAX_VALUE,
                        createTotal = 0,
                        createCount = 0,
                        createErrCount = 0;
                long start0 = System.currentTimeMillis();
                String ts = "" + new java.sql.Timestamp(System.currentTimeMillis()) + "> ";
                System.out.println(ts+ "Latency test...");
                for (int i = 0; i < 50000; i++) {
                    try {
                        long start = System.currentTimeMillis();
                        //
                        print("some message to print by MT4 Terminal");
                        //
                        long time = System.currentTimeMillis() - start;
                        //
                        getCounted(time, createCnt);
                        //
                        createMax = Math.max(createMax, time);
                        createMin = Math.min(createMin, time);
                        createCount++;
                    } catch (Exception e) {
                        createErrCount++;
                    } finally {
                        createTotal++;
                    }
                }
                long time0 = System.currentTimeMillis() - start0;
                ts = "" + new java.sql.Timestamp(System.currentTimeMillis()) + "> ";
                System.out.println(ts+ "total latency test duration: " + time0 + " millis for " + createTotal + " requests");
                System.out.println(format(ts,
                    "Total Req",
                        "Error Req",
                        "Min_Req Time",
                        "Max_Req Time",
                        "Avg_Req Time",
                        CNT_THRESHOLDS));
                System.out.println(format(ts,
                    "Cnt " + createCount,
                        "ErrCnt " + createErrCount,
                        "Min " + (createMin == Long.MAX_VALUE ? 0 : createMin),
                        "Max " + createMax,
                        "Avg " + ((int) ((double) createTotal / createCount)),
                        createCnt));
            }
            //
            int ordersCount = ordersTotal();
            //
            if (ordersCount == 0) {
                System.out.println("ordersTotal: " + ordersCount + ", accBal=" + accountBalance());
                //
                double point = marketInfo(getSymbol(), MarketInfo.MODE_POINT);
                double price = marketInfo(getSymbol(), MarketInfo.MODE_BID);
                sellPrice = price;
                int ticket = orderSend(
                        getSymbol(),
                        TradeOperation.OP_SELL,
                        1,
                        sellPrice,
                        2,
                        0,//price + 200 * point,
                        0,//price - 200 * point,
                        "" + System.currentTimeMillis(),
                        0,
                        new Date(System.currentTimeMillis() + 60 * 60 * 1000),
                        -1
                );
                System.out.println("---------------------------------------");
                System.out.println("Sell order ticket: " + ticket);
            } else {
                if (orderSelect(0, SelectionType.SELECT_BY_POS, SelectionPool.MODE_TRADES)) {
                    int ticket = orderTicket();
                    String orderSymbol = orderSymbol();
                    double bid = marketInfo(orderSymbol, MarketInfo.MODE_BID);
                    double ask = marketInfo(orderSymbol, MarketInfo.MODE_ASK);
                    //
                    if (!infoPrinted) {
                        long serverTime = ((long) marketInfo(orderSymbol(), MarketInfo.MODE_TIME)) * 1000;
                        long l = (serverTime - System.currentTimeMillis()) / 1000 / 60;
                        System.out.println("Found order, ticket=" + ticket + ", profit=" + orderProfit() + ", symbol=" + orderSymbol + ", volume=" + orderLots() + ", orderOpenTime=" + orderOpenTime());
                        if (getTermServerHost() != null && getTermServerHost().equals("127.0.0.1")) {
//                            alert("Found order, ticket=" + ticket + ", profit=" + orderProfit() + ", symbol=" + orderSymbol + ", volume=" + orderLots() + ", orderOpenTime=" + orderOpenTime());
                            //
                            String text = "Found order, ticket=" + ticket + ", profit=" + orderProfit() + ", symbol=" + orderSymbol + ", volume=" + orderLots() + ", orderOpenTime=" + orderOpenTime()
                                    + " minWindowPrice=" + windowPriceMin(0)
                                    + " maxWindowPrice=" + windowPriceMax(0);
                            String caption = terminalCompany();
                            int res = messageBox(
                                    text,
                                    caption,
                                    MessageBoxFlag._MB_OK | MessageBoxFlag._MB_ICONINFORMATION
                            );
                            if (MessageBoxID._IDOK == res) {
                                System.out.println("MessageBox result: IDOK");
                            }
                        }
                        //
                        infoPrinted = true;
                    }
                    //
                    String startTime = orderComment();
                    long _startTime = 0;
                    try {
                        _startTime = Long.parseLong(startTime);
                    } catch (NumberFormatException e) {
                    }
                    if (_startTime == 0) {
                        System.out.println("Closing order, ticket=" + ticket + ", profit=" + orderProfit() + ", symbol=" + orderSymbol + ", volume=" + orderLots() + ", orderOpenTime=" + orderOpenTime());
                        if (orderType() == TradeOperation._OP_BUY) {
                            orderClose(ticket, orderLots(), bid, 2, 0);
                        } else {
                            orderClose(ticket, orderLots(), marketInfo(orderSymbol, MarketInfo.MODE_ASK), 2, 0);
                        }
                    } else {
                        long seconds = (System.currentTimeMillis() - _startTime) / 1000;
                        if (seconds % 100 == 0) {
                            if (orderProfit() > 1000) {
                                System.out.println("Closing order, ticket=" + ticket + ", profit=" + orderProfit() + ", orderOpenTime=" + orderOpenTime());
                                orderClose(ticket, 1, ask, 2, 0);
                            } else if (seconds > 6000) {
                                System.out.println("Closing order, ticket=" + ticket + ", loss=" + orderProfit() + ", orderOpenTime=" + orderOpenTime());
                                orderClose(ticket, 1, ask, 2, 0);
                            }
                        } else {
                            if (shift++ < 10000) {
                                //
                                // Load history
                                //
                                try {
                                    iOpen(getSymbol(), Timeframe.PERIOD_H4, shift);
                                    iClose(getSymbol(), Timeframe.PERIOD_H4, shift);
                                    iHigh(getSymbol(), Timeframe.PERIOD_H4, shift);
                                    iLow(getSymbol(), Timeframe.PERIOD_H4, shift);
                                    if (shift % 10 == 0) {
                                        System.out.println("History H1 shift: " + shift);
                                    }
                                } catch (Throwable ignore) {
                                }
                            }
                        }
                    }
                }
            }
        } catch (ErrCustomIndicatorError errCustomIndicatorError) {
            errCustomIndicatorError.printStackTrace();
        } catch (ErrIntegerParameterExpected errIntegerParameterExpected) {
            errIntegerParameterExpected.printStackTrace();
        } catch (ErrInvalidFunctionParamvalue errInvalidFunctionParamvalue) {
            errInvalidFunctionParamvalue.printStackTrace();
        } catch (ErrInvalidPriceParam errInvalidPriceParam) {
            errInvalidPriceParam.printStackTrace();
        } catch (ErrInvalidTicket errInvalidTicket) {
            errInvalidTicket.printStackTrace();
        } catch (ErrUnknownSymbol errUnknownSymbol) {
            errUnknownSymbol.printStackTrace();
        } catch (ErrTradeNotAllowed errTradeNotAllowed) {
            errTradeNotAllowed.printStackTrace();
        } catch (ErrStringParameterExpected errStringParameterExpected) {
            errStringParameterExpected.printStackTrace();
        } catch (ErrShortsNotAllowed errShortsNotAllowed) {
            errShortsNotAllowed.printStackTrace();
        } catch (ErrLongsNotAllowed errLongsNotAllowed) {
            errLongsNotAllowed.printStackTrace();
        } catch (ErrNoOrderSelected errNoOrderSelected) {
            errNoOrderSelected.printStackTrace();
//        } catch (ErrHistoryWillUpdated errHistoryWillUpdated) {
//            System.out.println("Requested history data in updating state.");
        } catch (ErrNotEnoughMoney errNotEnoughMoney) {
            throw new RuntimeException(errNotEnoughMoney);
        } catch (ErrCommonError errCommonError) {
            throw new RuntimeException(errCommonError);
        } catch (ErrPriceChanged errPriceChanged) {
            throw new RuntimeException(errPriceChanged);
        } catch (ErrLongPositionsOnlyAllowed errLongPositionsOnlyAllowed) {
            throw new RuntimeException(errLongPositionsOnlyAllowed);
        } catch (ErrInvalidPrice errInvalidPrice) {
            throw new RuntimeException(errInvalidPrice);
        } catch (ErrServerBusy errServerBusy) {
            throw new RuntimeException(errServerBusy);
        } catch (ErrOldVersion errOldVersion) {
            throw new RuntimeException(errOldVersion);
        } catch (ErrOffQuotes errOffQuotes) {
            throw new RuntimeException(errOffQuotes);
        } catch (ErrTradeTimeout4 errTradeTimeout4) {
            throw new RuntimeException(errTradeTimeout4);
        } catch (ErrTooFrequentRequests errTooFrequentRequests) {
            throw new RuntimeException(errTooFrequentRequests);
        } catch (ErrInvalidStops errInvalidStops) {
            throw new RuntimeException(errInvalidStops);
        } catch (ErrTradeContextBusy errTradeContextBusy) {
            throw new RuntimeException(errTradeContextBusy);
        } catch (ErrNoConnection errNoConnection) {
            throw new RuntimeException(errNoConnection);
        } catch (ErrInvalidAccount errInvalidAccount) {
            throw new RuntimeException(errInvalidAccount);
        } catch (ErrTooManyRequests errTooManyRequests) {
            throw new RuntimeException(errTooManyRequests);
        } catch (ErrTradeTimeout2 errTradeTimeout2) {
            throw new RuntimeException(errTradeTimeout2);
        } catch (ErrInvalidTradeParameters errInvalidTradeParameters) {
            throw new RuntimeException(errInvalidTradeParameters);
        } catch (ErrTradeTimeout3 errTradeTimeout3) {
            throw new RuntimeException(errTradeTimeout3);
        } catch (ErrOrderLocked errOrderLocked) {
            throw new RuntimeException(errOrderLocked);
        } catch (ErrMarketClosed errMarketClosed) {
            throw new RuntimeException(errMarketClosed);
        } catch (ErrRequote errRequote) {
            throw new RuntimeException(errRequote);
        } catch (ErrInvalidTradeVolume errInvalidTradeVolume) {
            throw new RuntimeException(errInvalidTradeVolume);
        } catch (ErrAccountDisabled errAccountDisabled) {
            throw new RuntimeException(errAccountDisabled);
        } catch (ErrTradeTimeout errTradeTimeout) {
            throw new RuntimeException(errTradeTimeout);
        } catch (ErrTradeTooManyOrders errTradeTooManyOrders) {
            throw new RuntimeException(errTradeTooManyOrders);
        } catch (ErrTradeModifyDenied errTradeModifyDenied) {
            throw new RuntimeException(errTradeModifyDenied);
        } catch (ErrTradeExpirationDenied errTradeExpirationDenied) {
            throw new RuntimeException(errTradeExpirationDenied);
        } catch (ErrTradeDisabled errTradeDisabled) {
            throw new RuntimeException(errTradeDisabled);
        }
    }

    public static void main(String[] args) throws IOException, InterruptedException {
        if (args.length == 1 && args[0].equals("test")) {
            JFXServer jfxServer = new JFXServer("localhost", 7777);
            //
            System.out.println("Working...");
            Thread.sleep(2000000);
            //
            System.out.println("Stopping JFXServer...");
            JFXServer.stop();
            //
            System.out.println("Exit.");
        } else {
            _main(args);
        }
    }

    public static void _main(String[] args) throws IOException, InterruptedException {
        final JFXExample jfxExample = new JFXExample();
        //
        Runtime.getRuntime().addShutdownHook(new Thread() {
            public void run() {
                try {
                    System.out.println("MT4 Terminal client disconnect...");
                    jfxExample.disconnect();
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }
        });
        //
        System.out.println("You may kill terminal.exe to test reconnection...");
        jfxExample.setReconnect(true);
        //
        String terminalServerIpAddress = "127.0.0.1";

        jfxExample.connect(terminalServerIpAddress, 7788, new Broker("VPEAG-Demo"), "119968", "a5gndtm");

//        jfxExample.connect(terminalServerIpAddress, 7788, Broker.MBTrading_Demo, "500433241", "6qjjovx");
//        jfxExample.connect(terminalServerIpAddress, 7788, Broker.MBTrading_Demo, "500432727", "n7tkcxi");

//        jfxExample.connect(terminalServerIpAddress, 7788, Broker.AlpariUS_Demo, "982388", "gtx8TCo");
//        jfxExample.connect(terminalServerIpAddress, 7788, new Broker("AlpariUS-Demo"), "982388", "gtx8TCo");
        //
        System.out.println("Working...");
        Thread.sleep(1200000);
        //
        System.out.println("Stopping JFXServer...");
        JFXServer.stop();
        //
        System.out.println("Exit.");
    }
}
